/**
Action Script library. Key class implementation
Copyright (C) 2005 

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_key.hh"
#include "../lib/event_manager.hh"

typedef void (ASKey::*funcPtrKey)();
static funcPtrKey callbacks[]= { 
	&ASKey::init,			// 0
	&ASKey::addListener,	// 1
	&ASKey::getAscii,		// 2
	&ASKey::getCode,		// 3
	&ASKey::isDown,			// 4
	&ASKey::isToggled,		// 5
	&ASKey::removeListener,	// 6
	&ASKey::onKeyDown,		// 7
	&ASKey::onKeyUp,		// 8
	// Need to be at the end.
	&ASKey::addProperty,		
	&ASKey::registerClass,	
	&ASKey::AStoString,		
	&ASKey::unwatch,			
	&ASKey::valueOf,			
	&ASKey::watch			
};

/**
 * The constructor
 **/
ASKey::ASKey(Script *s, Graphics *g) : ASObject(s,g,L"Key") {
}

/**
 * The destructor
 **/
ASKey::~ASKey() {
}

/**
 * Return a function descriptor
 */
Function *ASKey::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"Key") 			{func->setFunctionIndex(0); func->setNumParams(0); }
	if (memberName == L"addListener") 	{func->setFunctionIndex(1); func->setNumParams(1); }
	if (memberName == L"getAscii") 		{func->setFunctionIndex(2); func->setNumParams(0); }
	if (memberName == L"getCode") 		{func->setFunctionIndex(3); func->setNumParams(0); }
	if (memberName == L"isDown") 		{func->setFunctionIndex(4); func->setNumParams(1); }
	if (memberName == L"isToggled") 	{func->setFunctionIndex(5); func->setNumParams(1); }
	if (memberName == L"removeListener"){func->setFunctionIndex(6); func->setNumParams(1); }
	if (memberName == L"onKeyDown")		{func->setFunctionIndex(7); func->setNumParams(1); }
	if (memberName == L"onKeyUp")		{func->setFunctionIndex(8); func->setNumParams(1); }
	func = finishFunctionSetup(memberName,9,func);
	return func;
}

/**
 * Call class built in functions
 **/
void ASKey::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}

/**
 *
 **/
void ASKey::init() {
	puts("TODO - ASKey::init()");
}

/**
 * Registers an object to receive notification when the onKeyDown and onKeyUp methods are invoked.
 **/
void ASKey::addListener() {
	if (numPassedParams > 0) {
		ASObject * a = (ASObject*)script->getRegister(1);
		// TODO: later check for getMember() in addition to getFunctionMember?
		if (a->getFunctionMember(L"onKeyDown") != NULL) listeners[L"onKeyDown"] = a; 
		if (a->getFunctionMember(L"onKeyUp") != NULL) listeners[L"onKeyUp"] = a;
	}
	noReturnValue();
}

/**
 * Returns the ASCII value of the last key pressed.
 **/
void ASKey::getAscii() {
	puts("TODO - ASKey::getAscii()");
	noReturnValue();
}

/**
 * Returns the virtual key code of the last key pressed.
 **/
void ASKey::getCode() {
	script->push(new IntegerObject(script->getEventManager()->getLastKeyPressed()));
}

/**
 * Returns true if the key specified in the parameter is pressed.
 **/
void ASKey::isDown() {
	bool res = false;
	if (numPassedParams > 0) {
		int key = int(script->getRegister(1)->toNumeric());
		res = script->getEventManager()->getLastKeyPressed() == key;
	} 
	script->push(new BooleanObject(res));
}

/**
 * Returns true if the Num Lock or Caps Lock key is activated.
 **/
void ASKey::isToggled() {
	puts("TODO - ASKey::isToggled()");
	noReturnValue();
}

/**
 *  Removes an object that was previously registered with Key.addListener().
 **/
void ASKey::removeListener() {
	puts("TODO - ASKey::removeListener()");
	noReturnValue();
}

void ASKey::onKeyDown() {
	ASObject *listener = getListener(L"onKeyDown");
	if (listener != NULL)
	{
		Function *f = listener->getFunctionMember(L"onKeyDown");
		listeners[L"onKeyDown"]->callFunction(f);
	}

	noReturnValue();
}

void ASKey::onKeyUp(){
	ASObject *listener = getListener(L"onKeyUp");
	if (listener != NULL)
	{
		Function *f = listener->getFunctionMember(L"onKeyUp");
		listeners[L"onKeyUp"]->callFunction(f);
	}

	noReturnValue();
}

/**
 * Return a member
 **/
ValueObject*ASKey::getMember(wstring memberName) {
	wstring name = toUpper(memberName);
	ValueObject *res = NULL;
	if (name == L"BACKSPACE") {
		res = new IntegerObject(8);
	} else if (name == L"CAPSLOCK") {
		res = new IntegerObject(20);
	} else if (name == L"CONTROL") {
		res = new IntegerObject(17);
	} else if (name == L"DELETEKEY") {
		res = new IntegerObject(46);
	} else if (name == L"DOWN") {
		res = new IntegerObject(40);
	} else if (name == L"END") {
		res = new IntegerObject(35);
	} else if (name == L"ENTER") {
		res = new IntegerObject(13);
	} else if (name == L"ESCAPE") {
		res = new IntegerObject(27);
	} else if (name == L"HOME") {
		res = new IntegerObject(36);
	} else if (name == L"INSERT") {
		res = new IntegerObject(45);
	} else if (name == L"LEFT") {
		res = new IntegerObject(37);
	} else if (name == L"PGDN") {
		res = new IntegerObject(34);
	} else if (name == L"PGUP") {
		res = new IntegerObject(33);
	} else if (name == L"RIGHT") {
		res = new IntegerObject(39);
	} else if (name == L"SHIFT") {
		res = new IntegerObject(16);
	} else if (name == L"SPACE") {
		res = new IntegerObject(32);
	} else if (name == L"TAB") {
		res = new IntegerObject(9);
	} else if (name == L"UP") {
		res = new IntegerObject(38);
	} else {
		res = defaultGetMember(memberName);
	}
	return res;
}
