/**
Action Script library. Action Script Linker class. - Built-in class linkage
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net) 

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_linker.hh"
#include "as_accessibility.hh"
#include "as_arguments.hh"
#include "as_array.hh"
#include "as_boolean.hh"
#include "as_button.hh"
#include "as_camera.hh"
#include "as_color.hh"
#include "as_context_menu.hh"
#include "as_context_menuitem.hh"
#include "as_custom_actions.hh"
#include "as_date.hh"
#include "as_error.hh"
#include "as_internal.hh"
#include "as_function.hh"
#include "as_key.hh"
#include "as_load_vars.hh"
#include "as_local_connection.hh"
#include "as_math.hh"
#include "as_microphone.hh"
#include "as_mouse.hh"
#include "as_movie_clip.hh"
#include "as_movie_clip_loader.hh"
#include "as_net_connection.hh"
#include "as_net_stream.hh"
#include "as_number.hh"
#include "as_object.hh"
#include "as_print_job.hh"
#include "as_selection.hh"
#include "as_shared_object.hh"
#include "as_sound.hh"
#include "as_stage.hh"
#include "as_string.hh"
#include "as_system.hh"
#include "as_system_capabilities.hh"
#include "as_system_security.hh"
#include "as_text_field.hh"
#include "as_text_field_stylesheet.hh"
#include "as_text_format.hh"
#include "as_text_snapshot.hh"
#include "as_video.hh"
#include "as_xml.hh"
#include "as_xml_node.hh"
#include "as_xml_socket.hh"
#include <iostream>
#include "../lib/swf_engine.hh"

extern bool debug;

/**
 * The constructor
 * 
 * The linker class is instanciated only once per flash instance, so the folowing code does not have a big performance impact.
 **/
ASLinker::ASLinker(Script *s, Graphics *g) {
	this->script = s;
	this->graphics = g;

	// Creates the className / class object relation
	classes[L"Accessibility"] = new ASAccessibility(s,g); 
	classes[L"Arguments"] = new ASArguments(s,g);
	classes[L"Array"] = new ASArray(s,g); 
	classes[L"Boolean"] = new ASBoolean(s,g); 
	classes[L"Button"] = new ASButton(s,g); 
	classes[L"Camera"] = new ASCamera(s,g); 
	classes[L"Color"] = new ASColor(s,g); 
	classes[L"ContextMenu"] = new ASContextMenu(s,g); 
	classes[L"ContextMenuItem"] = new ASContextMenuItem(s,g); 
	classes[L"CustomActions"] = new ASCustomActions(s,g); 
	classes[L"Date"] = new ASDate(s,g); 
	classes[L"Error"] = new ASError(s,g); 
	classes[L"Function"] = new ASFunction(s,g); 
	classes[L"Internal"] = new ASInternal(s,g); 
	classes[L"Key"] = new ASKey(s,g); 
	classes[L"LoadVars"] = new ASLoadVars(s,g); 
	classes[L"LocalConnection"] = new ASLocalConnection(s,g); 
	classes[L"Math"] = new ASMath(s,g);
	classes[L"math"] = new ASMath(s,g);
	classes[L"Microphone"] = new ASMicrophone(s,g); 
	classes[L"Mouse"] = new ASMouse(s,g); 
	classes[L"MovieClip"] = new ASMovieClip(s,g); 
	classes[L"MovieClipLoader"] = new ASMovieClipLoader(s,g); 
	classes[L"NetConnection"] = new ASNetConnection(s,g); 
	classes[L"NetStream"] = new ASNetStream(s,g); 
	classes[L"Number"] = new ASNumber(s,g); 
	classes[L"Object"] = new ASObject(s,g); 
	classes[L"PrintJob"] = new ASPrintJob(s,g); 
	classes[L"Selection"] = new ASSelection(s,g); 
	classes[L"SharedObject"] = new ASSharedObject(s,g); 
	classes[L"Sound"] = new ASSound(s,g); 
	classes[L"Sprite"] = new ASSprite(s,g); 
	classes[L"Stage"] = new ASStage(s,g); 
	classes[L"String"] = new ASString(s,g); 
	classes[L"System"] = new ASSystem(s,g); 
	classes[L"SystemCapabilities"] = new ASSystemCapabilities(s,g); 
	classes[L"SystemSecurity"] = new ASSystemSecurity(s,g); 
	classes[L"TextField"] = new ASTextField(s,g); 
	classes[L"TextFieldStylesheet"] = new ASTextFieldStyleSheet(s,g); 
	classes[L"TextFormat"] = new ASTextFormat(s,g); 
	classes[L"TextSnapshot"] = new ASTextSnapshot(s,g); 
	classes[L"Video"] = new ASVideo(s,g); 
	classes[L"XML"] = new ASXml(s,g); 
	classes[L"XMLNode"] = new ASXmlNode(s,g); 
	classes[L"XMLSocket"] = new ASXmlSocket(s,g);

	// creates the DirectFunctionCall / class reference
	callFunctions[L"attachMovie"] = L"MovieClip";
	callFunctions[L"beginFill"] = L"MovieClip";
	callFunctions[L"beginGradientFill"] = L"MovieClip";
	callFunctions[L"clear"] = L"MovieClip";
	callFunctions[L"createEmptyMovieClip"] = L"Sprite";
	callFunctions[L"createTextField"] = L"MovieClip";
	callFunctions[L"curveTo"] = L"MovieClip";
	callFunctions[L"duplicateMovieClip"] = L"MovieClip";
	callFunctions[L"enable"] = L"????????????????????????";
	callFunctions[L"endFiclasses[className];ll"] = L"MovieClip";
	callFunctions[L"focusEnabled"] = L"MovieClip";
	callFunctions[L"getBounds"] = L"MovieClip";
	callFunctions[L"getBytesLoaded"] = L"MovieClip";
	callFunctions[L"getBytesTotal"] = L"MovieClip";
	callFunctions[L"getDepth"] = L"MovieClip";
	callFunctions[L"getInstanceAtDepth"] = L"MovieClip";
	callFunctions[L"getNextHighestDepth"] = L"MovieClip";
	callFunctions[L"getSWFVersion"] = L"MovieClip";
	callFunctions[L"getTextSnapshot"] = L"MovieClip";
	callFunctions[L"getURL"] = L"MovieClip";
	callFunctions[L"globalToLocal"] = L"MovieClip";
	callFunctions[L"gotoAndPlay"] = L"MovieClip";
	callFunctions[L"gotoAndStop"] = L"MovieClip";
	callFunctions[L"hitArea"] = L"MovieClip";
	callFunctions[L"hitTest"] = L"MovieClip";
	callFunctions[L"lineStyle"] = L"MovieClip";
	callFunctions[L"lineTo"] = L"MovieClip";
	callFunctions[L"loadMovie"] = L"MovieClip";
	callFunctions[L"loadVariables"] = L"MovieClip";
	callFunctions[L"localToGlobal"] = L"MovieClip";
	callFunctions[L"menu"] = L"MovieClip";
	callFunctions[L"moveTo"] = L"MovieClip";
	callFunctions[L"nextFrame"] = L"MovieClip";
	callFunctions[L"play"] = L"MovieClip";
	callFunctions[L"prevFrame"] = L"MovieClip";
	callFunctions[L"removeMovieClip"] = L"MovieClip";
	callFunctions[L"setMask"] = L"MovieClip";
	callFunctions[L"setInterval"] = L"Internal";
	callFunctions[L"startDrag"] = L"MovieClip";
	callFunctions[L"stop"] = L"MovieClip";
	callFunctions[L"stopDrag"] = L"MovieClip";
	callFunctions[L"swapDepths"] = L"MovieClip";
	callFunctions[L"tabChildren"] = L"MovieClip";
	callFunctions[L"tabEnabled"] = L"MovieClip";
	callFunctions[L"tabIndex"] = L"MovieClip";
	callFunctions[L"trackAsMenu"] = L"MovieClip";
	callFunctions[L"unloadMovie"] = L"MovieClip";
	callFunctions[L"useHandCursor"] = L"MovieClip";
	callFunctions[L"ASSetPropFlags"] = L"Internal";

	callFunctions[L"parseInt"] = L"Function";
	callFunctions[L"clearInterval"] = L"Function";
	callFunctions[L"eval"] = L"Function";
	callFunctions[L"fscommand"] = L"Function";
	callFunctions[L"gotoAndPlay"] = L"Function";
	callFunctions[L"gotoAndStop"] = L"Function";
	callFunctions[L"isNaN"] = L"Function";
	callFunctions[L"loadMovieNum"] = L"Function";
	callFunctions[L"loadVariablesNum"] = L"Function";
	callFunctions[L"MMExecute"] = L"Function";
	callFunctions[L"nextScene"] = L"Function";
	callFunctions[L"on"] = L"Function";
	callFunctions[L"onClipEvent"] = L"Function";
	callFunctions[L"parseFloat"] = L"Function";
	callFunctions[L"parseInt"] = L"Function";
	callFunctions[L"print"] = L"Function";
	callFunctions[L"printAsBitmap"] = L"Function";
	callFunctions[L"printAsBitmapNum"] = L"Function";
	callFunctions[L"printNum"] = L"Function";
	callFunctions[L"setInterval"] = L"Function";
	callFunctions[L"setProperty"] = L"Function";
	callFunctions[L"stopAllSounds"] = L"Function";
	callFunctions[L"trace"] = L"Function";
	callFunctions[L"updateAfterEvent"] = L"Function";
}

ASLinker::~ASLinker() {
	classes.clear();
	callFunctions.clear();
}

/**
 * resolve method dependancy
 **/
Function *ASLinker::getFunctionMember(wstring className, wstring methodName, ScriptObject **sc) {
	if (debug) std::wcout<<L"Lookup: "<<className<<L":"<<methodName;
	if (methodName == L"prototype") methodName = className;	// The prototype is the className for builtin classes
	ASInterface *c = classes[className];
	if (c != NULL) return c->getFunctionMember(methodName);
	if (className == L"_root" || className == L"_global") {
		wstring s = callFunctions[methodName];
		if (s != L"") {
			c = classes[s];
			if (c != NULL) {
				*sc = c;		// Set new ScriptObject to new object (static members);
				return c->getFunctionMember(methodName);
			}
		}
	}
	if (c != NULL)	return c->getFunctionMember(methodName);
	else return NULL;
}

ASInterface* ASLinker::getANewClassInstance(Script *s, Graphics *g,wstring className) {
	ASInterface *res = classes[className]->getNewInstance(s,g);
	return res;
}

/**
 * Call a built in function
 **/
void ASLinker::callFunction(ScriptObject *sc, Function *f) {
	if (sc->getObjectType() != ScriptObjectType || sc->getMovieName() == L"_global") sc = classes[f->getClassName()];
	sc->callFunction(f);
}

/**
 * Executes a function. 
 * The function may be an ActionScript built-in function (such as parseInt), a user-defined ActionScript function, or a native function.
 **/
Function *ASLinker::getCallFunction(wstring functionName,ScriptObject **sc) {
	Function *func = NULL;
	wstring s = callFunctions[functionName];
	
	// Check if we call a drawing function on a Sprite
	if (script->getEngine()->isSpriteObject((ASInterface*)*sc) && s == L"MovieClip") {
		// Search for a MovieClipObject
		ASSprite *sprite= (ASSprite*)*sc;
		for (IntASObjectMap::iterator it=sprite->getCharacters()->begin(); it != sprite->getCharacters()->end(); it++) {
			ASObject *pobj = (ASObject*)(*it).second;
			if (pobj->getClassName() == L"MovieClip") {
				*sc = pobj;
				break; 
			}
		}		
		func = classes[s]->getFunctionMember(functionName);
	} else if (s != L"") {
		func = classes[s]->getFunctionMember(functionName);
		*sc = classes[s];
	}
	return func;
}

/**
 * Get a class from static member
 * It's used when calling a static class method for example
 * If it's a builtin function, we return a new Class Instance.
 **/
ScriptObject *ASLinker::getClassInstance(wstring className) {
	ASInterface *c = classes[className];
	if (c != NULL) c = c->getNewInstance(script,graphics);
	return c;
}

/**
 * Get a single class
 **/
ScriptObject *ASLinker::getClassSingleton(wstring className) {
	ASInterface *c = classes[className];
	return c;
}
