/**
Action Script library. Sound class
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#ifndef _AS_SOUND_HH
#define _AS_SOUND_HH
#include "as_object.hh"
#include "errno.h"
#include "../lib/script.hh"
#include <ffmpeg/avcodec.h>
#include <ffmpeg/avformat.h>
#include <SDL/SDL.h>
#include <SDL/SDL_audio.h>
#include <queue>
#include <pthread.h>
#include "assert.h"


// Used to send arguments to processSoundStreamBlock from decodeSoundStreamBlock 

typedef struct {
	void* timer;
	void* sound;
	void* tag;
} PRARGS ;

class SoundBlock {
public:
	uint8_t *buffer;
	Uint32 dpos;
	int dlen;
	int frame;                // The frame to correspond streaming playback with video
	int skip;
	~SoundBlock(){ free(buffer); } 
};

class Adpcm {
	long	stereo;	// Destination format - note we always decompress to 16 bit
	int		nBits;  // number of bits in each sample
	long	valpred[2]; // Current state
	int		index[2];
	long	nSamples; // number of samples decompressed so far

	// Parsing Info
	unsigned char 	*src;
	long	bitBuf; // this should always contain at least 24 bits of data
	int		bitPos;

	void fillBuffer();
	long getBits(int n);
	long getSBits(int n);

public:
	Adpcm(unsigned char *buffer, long isStereo);
	int decompress(short * dst, long n); // return number of good samples
};

class SOUNDTRACK {
public:
	UI8		soundFormat;
	UI8		soundRate;
	bool	soundSize;
	bool	soundType;
	UI32	soundSampleCount;
	UI8*	soundData;
	UI32	soundDataSize;
	int		headPosition, tailPosition;
	AVFormatContext *pFormatCtx;
	AVCodec         *pCodec;
	AVCodecContext  *pCodecCtx;
	list<SoundBlock*>outputBuffers;
	Adpcm	*adpcmDecoder;
	UI8* 	ptrData;
	bool	isSoundStreamBlock;
};

/**
 * This is used to send audio from processSoundStreamBloc to fillAudioHandler
 */
class AudioQueues {
public:
	AudioQueues();
	AudioQueues(SwfTimer*);
	~AudioQueues();
	vector<UI16> *getSpriteIds();
	void newQueue(UI16 spriteId);
	void push(UI16 spriteId, SoundBlock* sb);
	void pop(UI16 spriteId);
	SoundBlock* front(UI16 spriteId);
	void erase(UI16 spriteId);
	void clear();
	bool empty();
	bool empty(UI16 curSpriteId);
	int size();
	int size(UI16 spriteId);
	void signal();
private:
	void lockQueue();
	void unlockQueue();
	SwfTimer* timer;
	int queueLocks;
	map<UI16,queue<SoundBlock*> >	audioQueues;	// This is where the audio is
};

typedef stdext::hash_map<int, SOUNDTRACK*, HashInt,  HashIntCompare> IntSOUNDTRACKHashMap;

class ASSound : public ASObject {
public:
	ASSound(Script *s, Graphics *g);
	~ASSound();
	Function *getFunctionMember(wstring memberName);
	void 	callFunction(Function *f);
	void 	init();
	void 	init(SoundStreamHeadTag*);
	ASInterface* getNewInstance(Script *s,Graphics *g) { return new ASSound(s,g); }

	void	attachSound();
	void	getBytesLoaded() ;
	void	getBytesTotal();
	void	getPan();
	void	getTransform();
	void	getVolume();
	void	loadSound();
	void	setPan();
	void	setTransform();
	void	setVolume();
	void	start();
	void	stop();
	bool	getSoundError(){ return soundError; }
	
	SwfTimer*	getTimer(){return timer;}

	// Set by processStream when all tags have been sent to ASSound
	void	setDecodingFinished(UI16 id, bool finish);
	// Tested by fillAudioHandler, will in case exit and signal processStream to exit
	bool	getDecodingFinished(UI16 id);

	// events
	void	onID3();
	void	onLoad();
	void	onSoundComplete();
/*	char	duration;
	char	ID3;
	char	position;
*/
	void	registerSoundTrack(DefineSoundTag *t);
	void	startSound(int soundId,SOUNDINFO *);
	IntSOUNDTRACKHashMap* getSoundTracks() { return &soundTracks; }
	SDL_AudioSpec *getAudioSpec() { return &audioSpec; }
	void	setSoundInProgress(bool p) { soundInProgress = p; }
	bool	getSoundInProgress() { return soundInProgress; }
	void	decodeNextBlock(SOUNDTRACK *s);
	void	processSoundStreamBloc(SoundStreamBlockTag *);
	void	playButtonSound(int soundId, SOUNDINFO *);
	AudioQueues* getAudioQueues(){ return audioQueues; }
	void	startSoundStream(SoundStreamBlockTag*);
private:
	UI8		playbackSoundRate;
	UI8		streamSoundCompression;
	UI8		streamSoundRate;
	UI16	streamSoundSampleCount;
	SI16	latencySeek;
	bool	playbackSoundSize;
	bool	playbackSoundType;
	bool	streamSoundSize;
	bool	streamSoundType;
	IntSOUNDTRACKHashMap soundTracks;
	SDL_AudioSpec audioSpec;
	bool	soundInProgress;
	bool	isLittleEndian;
	bool	soundOpened;
	bool	soundInitialized;
	map<int,DefineSoundTag*> soundTags;
	bool	soundStreaming;					// True while playing streaming sound. Set by processStreamingSound. Tested by decodeSoundStreamBlock
	int frameNumber;					// The number of the soundBlock playing. Tested by swf_engine
	struct timeval frameTime;                        	// Time when soundFrame was set
	SwfTimer*	timer;
	AudioQueues*	audioQueues;	// Taking audio from decodeSoundStreamBlock fillAudioHandler
	map<UI16,bool>	*decodingFinished;			// Set by processSoundStreamBlock
	int	decodeSoundTrack(SOUNDTRACK *s, bool hasDecoder);
	int*	purgeStream(const pair<UI16,queue<SoundBlock>* >);
	pthread_attr_t	pthread_attr;				// Attributes for streaming threads
	bool	openSound();					// Opens the audio device
	bool	soundError;			// True if sound non-functioning. Set and tested in init() and openSound()
};

#endif
