/**
Macromedia Flash. Object type implementation.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "object_types.hh"

#include <config.h> // to check if WORDS_BIGENDIAN is defined
 
float GETFP32(uint8_t *p){ 
	union  {
		float f;
		uint8_t u[4];
	} data;
#if !defined(WORDS_BIGENDIAN) // little endian
	data.u[0] = p[0];
	data.u[1] = p[1];
	data.u[2] = p[2];
	data.u[3] = p[3];
#else // big endian
	data.u[0] = p[3];
	data.u[1] = p[2];
	data.u[2] = p[1];
	data.u[3] = p[0];
#endif
	return data.f;
}

double GETFP64(uint8_t *p) {
	union {
		double d;
		uint8_t u[8];
	} data;

#if !defined(WORDS_BIGENDIAN) // little endian
	data.u[0] = p[4];
	data.u[1] = p[5];
	data.u[2] = p[6];
	data.u[3] = p[7];
	data.u[4] = p[0];
	data.u[5] = p[1];
	data.u[6] = p[2];
	data.u[7] = p[3];
#else // big endian
	data.u[0] = p[3];
	data.u[1] = p[2];
	data.u[2] = p[1];
	data.u[3] = p[0];
	data.u[4] = p[7];
	data.u[5] = p[6];
	data.u[6] = p[5];
	data.u[7] = p[4];
#endif
	return data.d;
}

int32_t GETINT32(uint8_t *p) {
	union  {
		int32_t i;
		uint8_t u[4];
	} data;
#if !defined(WORDS_BIGENDIAN) // little endian
	data.u[0] = p[0];
	data.u[1] = p[1];
	data.u[2] = p[2];
	data.u[3] = p[3];
#else // big endian
	data.u[0] = p[3];
	data.u[1] = p[2];
	data.u[2] = p[1];
	data.u[3] = p[0];
#endif  
	return data.i;
}

uint16_t GETUI16(uint8_t *p) {
	union {
		uint16_t i;
		uint8_t u[2];
	} data;
#if !defined(WORDS_BIGENDIAN) // little endian
	data.u[0] = p[0];
	data.u[1] = p[1];
#else // big endian
	data.u[0] = p[1];
	data.u[1] = p[0];
#endif
  return data.i;
}

int16_t GETINT16(uint8_t *p) {
	union {
		int16_t i; 
		uint8_t u[2];
	} data;
#if !defined(WORDS_BIGENDIAN) // little endian
	data.u[0] = p[0];
	data.u[1] = p[1];
#else // big endian
	data.u[0] = p[1];
	data.u[1] = p[0];
#endif
	return data.i;
}

/**
 * Base class Implementation
 **/
ValueObject::ValueObject() { objectType = ValueObjectType; }
ValueObject::ValueObject(char *p) { objectType = ValueObjectType; }
ValueObject::~ValueObject() {};
uint8_t ValueObject::getObjectType() { return objectType; }
int ValueObject::decode(uint8_t *) { return 0;};
void *ValueObject::getValue() { return NULL; }
ValueObject* ValueObject::add(ValueObject *p) { return NULL;};
ValueObject* ValueObject::sub(ValueObject *p) { return NULL;};
ValueObject* ValueObject::mul(ValueObject *p) { return NULL;};
double ValueObject::toNumeric() { return 0; }
ValueObject* ValueObject::div(ValueObject *p) { return NULL;};
ValueObject* ValueObject::equal(ValueObject *p){ return new ValueObject();};
ValueObject* ValueObject::less(ValueObject *p) { return new ValueObject();};
ValueObject* ValueObject::andl(ValueObject *p) { return new ValueObject();};
ValueObject* ValueObject::orl(ValueObject *p)  { return new ValueObject();};
ValueObject* ValueObject::notl() 			  { return new ValueObject();};
ValueObject* ValueObject::stringEquals(ValueObject *p) { return new ValueObject();};
int  ValueObject::length() { return 0;};
wstring ValueObject::getObjectTypeName() const { return L""; }
wstring ValueObject::toString() const { return wstring(L"null"); }
ValueObject* ValueObject::clone() { return new ValueObject(*this); }
void	ValueObject::setThisObjectInstanceName(wstring p) { objectInstanceName = p ; }
wstring	ValueObject::getThisObjectInstanceName() { return objectInstanceName ; }


/**
 * boolean value class implementation
 **/
BooleanObject::BooleanObject() {
	objectType = BooleanObjectType;
	objectInstanceName = L"_$$$Boolean";
}
BooleanObject::BooleanObject(bool p) {
	objectType = BooleanObjectType;
	value = p;
	objectInstanceName = L"_$$$Boolean";
}
int BooleanObject::decode(uint8_t *p) { 
	value = *p != 0; 
	return 1;
}
bool BooleanObject::getValue() { return value; }
double BooleanObject::toNumeric() { return value ? 1.0 : 0.0; }
ValueObject* BooleanObject::equal(ValueObject *p){ return new BooleanObject(this->toNumeric() == p->toNumeric());  };
ValueObject* BooleanObject::less(ValueObject *p) { return new BooleanObject(this->toNumeric() < p->toNumeric());  };
ValueObject* BooleanObject::andl(ValueObject *p) { return new BooleanObject(this->toNumeric()!=0.0 && p->toNumeric() != 0.0 );  };
ValueObject* BooleanObject::orl(ValueObject *p)  { return new BooleanObject(this->toNumeric()!=0.0 || p->toNumeric() != 0.0 );  };
ValueObject* BooleanObject::notl() 			  { return new BooleanObject(this->toNumeric()==0.0 );  };
wstring BooleanObject::getObjectTypeName() const { return L"Bool"; }
wstring BooleanObject::toString() const { return value ? wstring(L"true"): wstring(L"false"); }
BooleanObject* BooleanObject::clone() { return new BooleanObject(*this); }


/**
 *  String class implementation.
 **/
StringObject::StringObject() {
	objectType = StringObjectType;
	objectInstanceName = L"_$$$String";
}
StringObject::StringObject(wstring p) {
	objectType = StringObjectType;
	value = p;
	objectInstanceName = L"_$$$String";
}
StringObject::~StringObject() {
}
int StringObject::decode(uint8_t *p) { 
	value = utf8Towstring((char*)p); 
	return strlen((const char*)p)+1;
}
wstring StringObject::getValue() { return value; }
ValueObject* StringObject::add(ValueObject *p) { return new StringObject(value + p->toString());	}
ValueObject* StringObject::equal(ValueObject *p){ return new BooleanObject(this->toNumeric() == p->toNumeric());  };
ValueObject* StringObject::less(ValueObject *p) { return new BooleanObject(this->toNumeric() < p->toNumeric());  };
ValueObject* StringObject::andl(ValueObject *p) { return new BooleanObject(this->toNumeric()!=0.0 && p->toNumeric() != 0.0 );  };
ValueObject* StringObject::orl(ValueObject *p)  { return new BooleanObject(this->toNumeric()!=0.0 || p->toNumeric() != 0.0 );  };
ValueObject* StringObject::notl() 			  { return new BooleanObject(this->toNumeric()==0.0 );  };
ValueObject* StringObject::stringEquals(ValueObject *p) { 
	if (objectType == p->getObjectType()) return new BooleanObject( value == ((StringObject*)p)->getValue());
	else return new BooleanObject(false);
}
int StringObject::length() { return value.size(); }
void StringObject::substring(int index, int count) {
	if (index>=0 && index < (int)value.size() && count > 0) value = value.substr(index,count);
	else value = L"";
}
bool StringObject::stringLess(StringObject *p) { return this->value < p->getValue(); }
wstring StringObject::getObjectTypeName() const { return L"String"; }
wstring StringObject::toString() const { return value; }
StringObject* StringObject::clone() {	return new StringObject(this->value); }
double	StringObject::toNumeric() {
	float a;
	int i = swscanf(value.c_str(),L"%f",&a);
	if (i>0) return double(a); else return 0.0;
}


/**
 * 32-bit IEEE single-precision fp value implementation.
 **/
FloatObject::FloatObject() {
	objectType = FloatObjectType;
	objectInstanceName = L"_$$$Number";
}
FloatObject::FloatObject(float f) {
	objectType = FloatObjectType;
	value = f;
	objectInstanceName = L"_$$$Number";
}
int FloatObject::decode(uint8_t *p) { 
	value = GETFP32(p);
	return 4;
}
float FloatObject::getValue() { return value; }
ValueObject* FloatObject::add(ValueObject *p) { return new FloatObject(float(value + p->toNumeric())); }
ValueObject* FloatObject::sub(ValueObject *p) { return new FloatObject(float(value - p->toNumeric())); }
ValueObject* FloatObject::mul(ValueObject *p) { return new FloatObject(float(value * p->toNumeric())); }
ValueObject* FloatObject::div(ValueObject *p) { 
	float dividend = p->toNumeric();
	if (dividend != 0.0f) {
		value = value / dividend;	
		return new FloatObject(value); 
	} else  return new StringObject(wstring(L"NaN"));
}
double FloatObject::toNumeric() { return value; }
ValueObject* FloatObject::equal(ValueObject *p){ return new BooleanObject(this->toNumeric() == p->toNumeric());  };
ValueObject* FloatObject::less(ValueObject *p) { return new BooleanObject(this->toNumeric() < p->toNumeric());  };
ValueObject* FloatObject::andl(ValueObject *p) { return new BooleanObject(this->toNumeric()!=0.0 && p->toNumeric() != 0.0 );  };
ValueObject* FloatObject::orl(ValueObject *p)  { return new BooleanObject(this->toNumeric()!=0.0 || p->toNumeric() != 0.0 );  };
ValueObject* FloatObject::notl() 			  { return new BooleanObject(this->toNumeric()==0.0 );  };
wstring FloatObject::getObjectTypeName() const { return L"Float"; }
	wstring FloatObject::toString() const { 
	char tmp[100]; 
	if (fabs(value) < 10e14) sprintf(tmp,"%f",value); else sprintf(tmp,"%e",value);
	// Remove trailing 0
	bool end = false;
	for (int i= strlen(tmp)-1;i>0 && !end; i--) if (tmp[i] == '0') tmp[i] = 0; else end = true;
	if (tmp[strlen(tmp)-1] == '.') tmp[strlen(tmp)-1]=0;
	return stringTowstring(tmp); 
}
FloatObject* FloatObject::clone() { return new FloatObject(*this); }


/**
 * 32-bit integer class implementation.
 **/
IntegerObject::IntegerObject() {
	objectType = IntegerObjectType;
	objectInstanceName = L"_$$$Number";
}
IntegerObject::IntegerObject(int p) {
	objectType = IntegerObjectType;
	value = (int32_t)p;
	objectInstanceName = L"_$$$Number";
}
int IntegerObject::decode(uint8_t *p) { 
	value = GETINT32(p); 
	return 4;
}
int32_t IntegerObject::getValue() { return value; }
ValueObject* IntegerObject::add(ValueObject *p) { return new FloatObject(float(value + p->toNumeric())); }
ValueObject* IntegerObject::sub(ValueObject *p) { return new FloatObject(float(value - p->toNumeric())); }
ValueObject* IntegerObject::mul(ValueObject *p) { return new FloatObject(float(value * p->toNumeric())); }
ValueObject* IntegerObject::div(ValueObject *p) { 
	float dividend = p->toNumeric();
	if (dividend != 0.0f) {
		value = int (float(value) / dividend);	
		return new IntegerObject(value); 
	} else  return new StringObject(wstring(L"NaN"));
}
double IntegerObject::toNumeric() { return value; }	
ValueObject* IntegerObject::equal(ValueObject *p){ return new BooleanObject(this->toNumeric() == p->toNumeric());  };
ValueObject* IntegerObject::less(ValueObject *p) {  return new BooleanObject(this->toNumeric() < p->toNumeric());  };
ValueObject* IntegerObject::andl(ValueObject *p) { return new BooleanObject(this->toNumeric()!=0.0 && p->toNumeric() != 0.0 );  };
ValueObject* IntegerObject::orl(ValueObject *p)  { return new BooleanObject(this->toNumeric()!=0.0 || p->toNumeric() != 0.0 );  };
ValueObject* IntegerObject::notl() 			  { return new BooleanObject(this->toNumeric()==0.0 );  };
wstring IntegerObject::getObjectTypeName() const { return L"Int"; }
wstring IntegerObject::toString() const{ char tmp[100]; sprintf(tmp,"%d",value); return stringTowstring(tmp); }
IntegerObject* IntegerObject::clone() { return new IntegerObject(*this); }


/**
 * Register number class implementation
 **/
RegisterNumberObject::RegisterNumberObject() {
	objectType = RegisterNumberObjectType;
	valueRef = NULL;
	objectInstanceName = L"_$$$Number";
}
uint8_t RegisterNumberObject::getObjectType() { return (*valueRef)->getObjectType(); }
void RegisterNumberObject::bind(int index, ValueObject **p) {
	registerIndex = index;
	valueRef = p; 
}
ValueObject *RegisterNumberObject::getValue() { return *valueRef; }
double 		RegisterNumberObject::toNumeric() { return (*valueRef)->toNumeric(); }	
wstring RegisterNumberObject::getObjectTypeName() const { return L"r"; }
wstring 	RegisterNumberObject::toString() const { return (*valueRef)->toString(); }
RegisterNumberObject*RegisterNumberObject::clone() { return new RegisterNumberObject(*this); } 


/**
 * 64-bit IEEE double-precision class implementation.
 **/
DoubleObject::DoubleObject() {
	objectType = DoubleObjectType;
	objectInstanceName = L"_$$$Number";
}
DoubleObject::DoubleObject(double p) {
	objectType = DoubleObjectType;
	value = p;
	objectInstanceName = L"_$$$Number";
}
int DoubleObject::decode(uint8_t *p) { 
	value = GETFP64(p); 
	return 8;
}
double DoubleObject::getValue() { return value; }
ValueObject* DoubleObject::add(ValueObject *p) { return new FloatObject(float(value + p->toNumeric())); }
ValueObject* DoubleObject::sub(ValueObject *p) { return new FloatObject(float(value - p->toNumeric())); }
ValueObject* DoubleObject::mul(ValueObject *p) { return new FloatObject(float(value * p->toNumeric())); }
ValueObject* DoubleObject::div(ValueObject *p) { 
	float dividend = p->toNumeric();
	if (dividend != 0.0f) {
		value = value / dividend;	
		return new DoubleObject(value); 
	} else  return new StringObject(wstring(L"NaN"));
}
double DoubleObject::toNumeric() { return value; }	
ValueObject* DoubleObject::equal(ValueObject *p){ return new BooleanObject(this->toNumeric() == p->toNumeric());  };
ValueObject* DoubleObject::less(ValueObject *p) { return new BooleanObject(this->toNumeric() < p->toNumeric());  };
ValueObject* DoubleObject::andl(ValueObject *p) { return new BooleanObject(this->toNumeric()!=0.0 && p->toNumeric() != 0.0 );  };
ValueObject* DoubleObject::orl(ValueObject *p)  { return new BooleanObject(this->toNumeric()!=0.0 || p->toNumeric() != 0.0 );  };
ValueObject* DoubleObject::notl() 			  { return new BooleanObject(this->toNumeric()==0.0 );  };
wstring DoubleObject::getObjectTypeName()const  { return L"Double"; }
wstring DoubleObject::toString() const {
	char tmp[100]; 
	if (fabs(value) < 10e14) sprintf(tmp,"%f",value); else sprintf(tmp,"%e",value);
	// Remove trailing 0 and last .
	bool end = false;
	for (int i= strlen(tmp)-1;i>0 && !end; i--) if (tmp[i] == '0') tmp[i] = 0; else end = true;
	if (tmp[strlen(tmp)-1] == '.') tmp[strlen(tmp)-1]=0;
	return stringTowstring(tmp); 
}
DoubleObject* DoubleObject::clone() { return new DoubleObject(*this); }


/**
 * constant pool index (for indices < 256) class implementation.
 **/
Constant8Object::Constant8Object() {
	objectType = Constant8ObjectType;
	objectInstanceName = L"_$$$Number";
}
int Constant8Object::decode(uint8_t *p) {  return 0;	}
ValueObject*Constant8Object::getParam() { return value; }
int		Constant8Object::getPoolEntry() { return poolEntry; }
ValueObject  *Constant8Object::getValue() { return new StringObject(toString()); }
wstring Constant8Object::getObjectTypeName() const { return L"Cst8"; }
wstring Constant8Object::toString() const { char tmp[100]; sprintf(tmp,"%d",poolEntry); return stringTowstring(tmp); } 
Constant8Object* Constant8Object::clone() { return new Constant8Object(*this); }

/**
 * constant pool index (for indices >= 256) 
 **/
Constant16Object::Constant16Object() {
	objectType = Constant16ObjectType;
	objectInstanceName = L"_$$$Number";
}
int Constant16Object::decode(uint8_t *p) { return 0; }
ValueObject*Constant16Object::getParam() { return value; }
int		Constant16Object::getPoolEntry() { return poolEntry; }
ValueObject  *Constant16Object::getValue() { return new StringObject(toString()); }
wstring Constant16Object::getObjectTypeName() const { return L"Cst16"; }
wstring Constant16Object::toString() const { char tmp[100]; sprintf(tmp,"%d",poolEntry); return stringTowstring(tmp); }
Constant16Object* Constant16Object::clone() { return new Constant16Object(*this); }


/**
 * A stack pointer class implementation
 **/
StackPointerObject::StackPointerObject(UI8 *p) { 
	objectType = StackPointerObjectType;
	value = p;
}
UI8*	StackPointerObject::getValue() { return value; }


/**
 * The Object instance members class implementation.
 * 
 * The _global constains static members.
 **/
ObjectMembers::~ObjectMembers() {
	//puts("~ObjectMembers");
	for (StringValueObjectHashMap::const_iterator it=members.begin(); it != members.end(); it++) {
		ValueObject *v = (*it).second; 
		if (v == NULL)continue;
		if (v->getObjectType() != ScriptObjectType)	{
			delete v;
		}
	}
	members.clear();
}

StringValueObjectHashMap *ObjectMembers::getMembers() { return &members; }
void ObjectMembers::setMember(wstring memberName, ValueObject *v) {
	if (getMember(memberName) != NULL) {
		ValueObject *o = members[memberName];
		if (o->getObjectType() != ScriptObjectType) delete o;
	}
	if (v->getObjectType() != ScriptObjectType) members[memberName] = v->clone(); else members[memberName] = v;
}
void ObjectMembers::setMemberNoClone(wstring memberName, ValueObject *v) {
	if (getMember(memberName) != NULL) {
		ValueObject *o = members[memberName];
		if (o->getObjectType() != ScriptObjectType) delete o;
	}
	members[memberName] = v;
}

ValueObject* ObjectMembers::getMember(wstring memberName) {
	StringValueObjectHashMap::iterator it = members.find(memberName);

	if (it == members.end())
		return NULL;

	ValueObject *v = members[memberName];
	if (v->getObjectType() != ScriptObjectType) v = v->clone();

	return v;
}


/**
 * An ScriptObject (an instance of a class)
 **/
ScriptObject::ScriptObject(wstring objectInstanceName) {
	//printf("\nCreate SO: %x\n",this);
	objectType = ScriptObjectType;
	this->objectInstanceName = objectInstanceName;
	members = new ObjectMembers();
	shouldBeRemoved = false;
	shouldBeRemoved = false;
	parent = NULL;
	isASCreated = false;
}
ScriptObject::~ScriptObject() {
	//printf("\nDelete SO: %x\n",this);
	//wcout << "~ScriptObject " << movieName << L"\n";
	removeMembers();
}
wstring	ScriptObject::toString() const { return objectInstanceName; }
wstring ScriptObject::getObjectTypeName() const { return L"SO"; }
ScriptObject*ScriptObject::clone() {		// Don't clone complex objects
	return this;
}
ObjectMembers*ScriptObject::getMembers() { return members; }
void ScriptObject::setMember(wstring memberName, ValueObject *v) { defaultSetMember(memberName,v); }
void ScriptObject::setMemberNoClone(wstring memberName, ValueObject *v) { members->setMemberNoClone(memberName, v); }
ValueObject*ScriptObject::getMember(wstring memberName) { return defaultGetMember(memberName); }
void ScriptObject::callFunction(Function *func) { puts("ScriptObject::callFunction Not overloaded? - Bug!"); };
void ScriptObject::toGraphic() { puts("ScriptObject::toGraphic: To be overloaded!!!"); }
void		ScriptObject::setDirty() { needRedraw = true; }
bool		ScriptObject::isDirty() { return needRedraw; }
void ScriptObject::setMainMember(ValueObject  *v) { puts("ScriptObject::setMainMember Not overloaded- BUG!!!!!"); };
ValueObject*ScriptObject::getMainMember() { puts("ScriptObject::getMainMember Not overloaded- BUG!!!!!"); return NULL;};
ValueObject* ScriptObject::notl()	  { return new BooleanObject(toString() == L"undefined");  };
void		ScriptObject::setCharacterId(int p) { characterId = p; }
int			ScriptObject::getCharacterId() { return characterId; }
void 		ScriptObject::defaultSetMember(wstring memberName, ValueObject *v) { 
	members->setMember(memberName, v); 
	members->setMember(toUpper(memberName),v);
	if (v->getObjectType() == ScriptObjectType) {
		ScriptObject *t = (ScriptObject*)v;
		t->setParent(this);
	}
}

/**
 * If we query the class name, returns class
 * else: try by sensistive case search,
 *       if not found,try in uppercase
 **/
ValueObject*ScriptObject::defaultGetMember(wstring memberName) { 
	if (memberName == className) return this;
	ValueObject *v = members->getMember(memberName);
	if (v == NULL) v = members->getMember(toUpper(memberName));
	return v;
}

wstring		ScriptObject::getMovieName() { return movieName; }
void		ScriptObject::setMovieName(wstring s) { movieName = s; }
wstring		ScriptObject::getClassName() { return className; }
bool		ScriptObject::getShouldBeRemoved() { return shouldBeRemoved; }
void		ScriptObject::setShouldBeRemoved(bool p) { shouldBeRemoved = p; }
void		ScriptObject::removeMembers() {
	delete members;
	members = NULL;
}
ScriptObject*ScriptObject::getParent() { 
	return parent; 
}
void ScriptObject::setParent(ScriptObject*p) {
	parent = p;
}

bool ScriptObject::getIsASCreated() { return isASCreated; }
void ScriptObject::setASCreated(bool p) { isASCreated = p; }
void ScriptObject::setPrototype(ScriptObject* p ) { defaultSetMember(L"prototype",p); }
ScriptObject*ScriptObject::getPrototype() { return (ScriptObject*)defaultGetMember(L"prototype"); }

wstring	MEMBER_TOSTRING_NOCLONE(ScriptObject *o,wstring is) { 
	ValueObject *a = o->getMember(is); 
	wstring s = a->toString(); 
	delete a; 
	return s; 
}

float	MEMBER_TONUMERIC_NOCLONE(ScriptObject *o,wstring is) { 
	ValueObject *a = o->getMember(is); 
	float ret = a->toNumeric(); 
	delete a; 
	return ret; 
}
