/**
Macromedia Flash Object Types.
Copyright (C) 2004  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#ifndef _OBJECT_TYPES_HH
#define _OBJECT_TYPES_HH

#include "common.hh"
#include "swf.hh"
#include <math.h>
#include <iostream>
#include <map>

using namespace std;
class Function;

float    GETFP32(uint8_t *p);
double   GETFP64(uint8_t *p);
int32_t  GETINT32(uint8_t *p);
uint16_t GETUI16(uint8_t *p);
int16_t  GETINT16(uint8_t *p);
wstring	encodeCharacterName(int charId);

enum ValueObjectTypeEnum {
	ValueObjectType =1,
	StringObjectType = 2,
	FloatObjectType = 3,
	RegisterNumberObjectType = 4,
	BooleanObjectType = 5,
	DoubleObjectType = 6,
	IntegerObjectType = 7,
	Constant8ObjectType= 8,
	Constant16ObjectType= 9,
	StackPointerObjectType = 10,
	ScriptObjectType = 11
};


class BooleanObject;

/**
 * Base class
 **/
class ValueObject {
public:
	ValueObject();
	ValueObject(char *p);
	virtual ~ValueObject();
	virtual uint8_t getObjectType();
	virtual int decode(uint8_t *);
	void *getValue();
	virtual ValueObject* add(ValueObject *p);
	virtual ValueObject* sub(ValueObject *p);
	virtual ValueObject* mul(ValueObject *p);
	virtual double toNumeric();
	virtual ValueObject* div(ValueObject *p);
	virtual ValueObject* equal(ValueObject *p);
	virtual ValueObject* less(ValueObject *p);
	virtual ValueObject* andl(ValueObject *p);
	virtual ValueObject* orl(ValueObject *p);
	virtual ValueObject* notl();
	virtual ValueObject* stringEquals(ValueObject *p);
	virtual int  length();
	virtual	wstring getObjectTypeName() const;
	virtual wstring toString() const;
	virtual ValueObject* clone();
	void	setThisObjectInstanceName(wstring p);
	wstring	getThisObjectInstanceName();
protected:
	uint8_t	objectType;
	wstring	objectInstanceName;
private:
};

inline wostream &operator<<(wostream &s, const ValueObject&o) { s<<o.getObjectTypeName()<<":"<<o.toString(); return s; }


/**
 * boolean value
 **/
class BooleanObject : public ValueObject {
public:
	BooleanObject();
	BooleanObject(bool p);
	int decode(uint8_t *p);
	bool getValue();
	double toNumeric();
	ValueObject* equal(ValueObject *p);
	ValueObject* less(ValueObject *p);
	ValueObject* andl(ValueObject *p);
	ValueObject* orl(ValueObject *p);
	ValueObject* notl();
	wstring getObjectTypeName() const;
	wstring toString() const;
	BooleanObject* clone();
private:
	bool	value;
};

/**
 *  String
 **/
class StringObject : public ValueObject {
public:
	StringObject();
	StringObject(wstring p);
	~StringObject();
	int decode(uint8_t *p);
	wstring getValue();
	ValueObject* add(ValueObject *p);
	ValueObject* equal(ValueObject *p);
	ValueObject* less(ValueObject *p);
	ValueObject* andl(ValueObject *p);
	ValueObject* orl(ValueObject *p);
	ValueObject* notl();
	ValueObject* stringEquals(ValueObject *p);
	int length();
	void substring(int index, int count);
	bool stringLess(StringObject *p);
	wstring getObjectTypeName() const;
	wstring toString() const;
	StringObject* clone();
	double	toNumeric();
private:
	wstring	value;
};


/**
 * 32-bit IEEE single-precision fp value
 **/
class FloatObject : public ValueObject {
public:
	FloatObject();
	FloatObject(float f);
	int decode(uint8_t *p);
	float getValue();
	ValueObject* add(ValueObject *p);
	ValueObject* sub(ValueObject *p);
	ValueObject* mul(ValueObject *p);
	ValueObject* div(ValueObject *p);
	double toNumeric();
	ValueObject* equal(ValueObject *p);
	ValueObject* less(ValueObject *p);
	ValueObject* andl(ValueObject *p);
	ValueObject* orl(ValueObject *p);
	ValueObject* notl();
	wstring getObjectTypeName() const;
	wstring toString() const;
	FloatObject* clone();
private:
	float 	value;
};

/**
 * 32-bit integer
 **/
class IntegerObject : public ValueObject {
public:
	IntegerObject();
	IntegerObject(int p);
	int decode(uint8_t *p);
	int32_t getValue();
	ValueObject* add(ValueObject *p);
	ValueObject* sub(ValueObject *p);
	ValueObject* mul(ValueObject *p);
	ValueObject* div(ValueObject *p);
	double toNumeric();
	ValueObject* equal(ValueObject *p);
	ValueObject* less(ValueObject *p);
	ValueObject* andl(ValueObject *p);
	ValueObject* orl(ValueObject *p);
	ValueObject* notl();
	wstring getObjectTypeName() const;
	wstring toString() const;
	IntegerObject* clone();
private:
	int32_t	value;
};

/**
 * Register number
 **/
class RegisterNumberObject : public ValueObject {
public:
	RegisterNumberObject();
	uint8_t getObjectType();
	void bind(int index, ValueObject **p);
	ValueObject *getValue();
	double 		toNumeric();
	wstring getObjectTypeName() const;
	wstring 	toString() const;
	RegisterNumberObject*clone();
private:
	ValueObject **valueRef;
	int			registerIndex;
};


/**
 * 64-bit IEEE double-precision
 **/
class DoubleObject : public ValueObject {
public:
	DoubleObject();
	DoubleObject(double p);
	int decode(uint8_t *p);
	double getValue();
	ValueObject* add(ValueObject *p);
	ValueObject* sub(ValueObject *p);
	ValueObject* mul(ValueObject *p);
	ValueObject* div(ValueObject *p);
	double toNumeric();
	ValueObject* equal(ValueObject *p);
	ValueObject* less(ValueObject *p);
	ValueObject* andl(ValueObject *p);
	ValueObject* orl(ValueObject *p);
	ValueObject* notl();
	wstring getObjectTypeName()const;
	wstring toString() const;
	DoubleObject* clone();
private:
	double	value;
};


/**
 * constant pool index (for indices < 256)
 **/
class Constant8Object : public ValueObject {
public:
	Constant8Object();
	int decode(uint8_t *p);
	ValueObject*getParam();
	int		getPoolEntry();
	ValueObject  *getValue();
	wstring getObjectTypeName() const;
 	wstring toString() const;
	Constant8Object* clone();
private:
	uint8_t poolEntry;
	ValueObject* value;
};

/**
 * constant pool index (for indices >= 256) 
 **/
class Constant16Object : public ValueObject {
public:
	Constant16Object();
	int decode(uint8_t *p);
	ValueObject*getParam();
	int		getPoolEntry();
	ValueObject  *getValue();
	wstring getObjectTypeName() const ;
 	wstring toString() const;
	Constant16Object* clone();
private:
	uint16_t poolEntry;
	ValueObject* value;
};

/**
 * A stack pointer
 **/
class StackPointerObject : public ValueObject {
public:
	StackPointerObject(UI8 *p);
	UI8*	getValue();
private:
	UI8*	value;
};


typedef stdext::hash_map<wstring, ValueObject *, HashUTF8String,  HashUTF8StringCompare> StringValueObjectHashMap;

/**
 * The Object instance members
 * 
 * The _global constains static members.
 **/
class ObjectMembers {
public:
	~ObjectMembers();
	StringValueObjectHashMap *getMembers();
	void setMember(wstring memberName, ValueObject *v);
	void setMemberNoClone(wstring memberName, ValueObject *v);
	ValueObject* getMember(wstring memberName);
private:
	StringValueObjectHashMap members;
};


/**
 * An ScriptObject (an instance of a class)
 **/
class ScriptObject: public ValueObject {
public:
	ScriptObject(wstring objectInstanceName);
	virtual ~ScriptObject();
	virtual wstring	toString() const;
	wstring getObjectTypeName() const;
	ScriptObject*clone();
	ObjectMembers*getMembers();
	virtual void setMember(wstring memberName, ValueObject *v);
	virtual void setMemberNoClone(wstring memberName, ValueObject *v);
	virtual ValueObject*getMember(wstring memberName);
	void	removeMembers();
	virtual void callFunction(Function *func);
	virtual void toGraphic();
	void		setDirty();
	bool		isDirty();
	virtual void setMainMember(ValueObject  *v);
	virtual ValueObject*getMainMember();
	ValueObject* notl();
	void		setCharacterId(int p);
	int			getCharacterId();
	wstring		getMovieName();
	void		setMovieName(wstring s);
	wstring		getClassName();
	bool		getShouldBeRemoved();
	void		setShouldBeRemoved(bool p);
	ScriptObject*getParent();
	void 		setParent(ScriptObject*);
	bool		getIsASCreated();
	void		setASCreated(bool p);
	void		setPrototype(ScriptObject*);
	ScriptObject*getPrototype();
protected:
	bool		needRedraw;
	void 		defaultSetMember(wstring memberName, ValueObject *v);
	ValueObject*defaultGetMember(wstring memberName);
	int			characterId;
	wstring		className, movieName;
	ObjectMembers*members;
	ScriptObject*parent;
private:
	bool		shouldBeRemoved;
	bool		isASCreated;

};

wstring	MEMBER_TOSTRING_NOCLONE(ScriptObject *o,wstring s);
float	MEMBER_TONUMERIC_NOCLONE(ScriptObject *o,wstring s);

#endif
