/**
Macromedia Flash File common implementation.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "common.hh"
#include "swf.hh"

GRADIENT::GRADIENT() {
	gradientRecords = new list<GRADRECORD*>;
};

GRADIENT::GRADIENT(GRADIENT *p) {
	numGradients = p->numGradients;
	gradientRecords = new list<GRADRECORD*>;
	for (list<GRADRECORD*>::const_iterator it = p->gradientRecords->begin(); it != p->gradientRecords->end(); it++) {
		GRADRECORD *a = *it;
		gradientRecords->push_back(new GRADRECORD(*a));
	}
}

GRADIENT::~GRADIENT() { 
	list<GRADRECORD*>::const_iterator it;
	for (it = gradientRecords->begin(); it != gradientRecords->end(); it++) {
		delete (*it);
	}
	gradientRecords->clear();
	delete gradientRecords;
}

FILLSTYLE::~FILLSTYLE() { 
	if (gradient!= NULL) {
		delete gradient;
		gradient = NULL;
	}
}

ClipActionRecord::~ClipActionRecord() {
	if (code != NULL) delete []code;
}

ClipActions::~ClipActions() { 
	list<ClipActionRecord*>::iterator it;
	for (it = clipActionRecords.begin(); it != clipActionRecords.end(); it++) 	delete (*it);
	clipActionRecords.clear(); 
}

BUTTONCONDACTION::~BUTTONCONDACTION() { 
	if (code != NULL) delete []code;
}

MORPHGRADIENT::~MORPHGRADIENT() { 
	list<MORPHGRADRECORD*>::iterator it;
	for (it = gradientRecords.begin(); it != gradientRecords.end(); it++) 	delete (*it);
	gradientRecords.clear();	
}

MORPHFILLSTYLES::~MORPHFILLSTYLES() { 
	if (fillStyles != NULL) delete []fillStyles; 
}

StyleChangeRecord::StyleChangeRecord() { 
	setTypeFlag(StyleChangeRecordType); 
	fillStyles = NULL; 
	lineStyles = NULL; 
}

StyleChangeRecord::~StyleChangeRecord() { 
	if (fillStyles != NULL) {
		delete []fillStyles; 
	}
	if (lineStyles != NULL) delete []lineStyles;
}

ShapeRecord*StyleChangeRecord::clone() { 
	StyleChangeRecord *sr = new StyleChangeRecord(*this);
	if (stateNewStyles) {
		sr->fillStyles = new FILLSTYLE[numFillStyles];
		for (int i=0; i< numFillStyles; i++) {
			sr->fillStyles[i] = fillStyles[i];
			if (fillStyles[i].getGradient() != NULL) {
				GRADIENT *a = new GRADIENT(fillStyles[i].getGradient());
				sr->fillStyles[i].setGradient(a);
			}
		}
		sr->lineStyles = new LINESTYLE[numLineStyles];
		for (int i=0; i< numLineStyles; i++) sr->lineStyles[i] = lineStyles[i];
	}
	return sr;
}

StraightEdgeRecord::StraightEdgeRecord() { 
	setTypeFlag(StraightEdgeRecordType); 
}

ShapeRecord*StraightEdgeRecord::clone() { 
	return new StraightEdgeRecord(*this);
}

ShapeRecord*CurvedEdgeRecord::clone() { 
	return new CurvedEdgeRecord(*this);
}

Shape::Shape() { 
	isShapeWithStyle = false; 
	shapeRecords = new list<ShapeRecord*>; 
}

Shape::~Shape() {	
	list<ShapeRecord*>::iterator it;
	for (it = shapeRecords->begin(); it != NULL && it != shapeRecords->end(); it++) 
		delete (*it);	
	shapeRecords->clear();
	delete shapeRecords;
}

Shape::Shape(Shape *p) {
	numFillBits = p->numFillBits;
	numLineBits = p->numLineBits;
	isShapeWithStyle = p->isShapeWithStyle;
	shapeRecords = new list<ShapeRecord*>;
	for (list<ShapeRecord*>::const_iterator i=p->getShapeRecords()->begin(); i != p->getShapeRecords()->end(); ++i) {
		shapeRecords->push_back((*i)->clone());
	}
}

ShapeWithStyle::ShapeWithStyle() { 
	isShapeWithStyle = true; 
	fillStyles = NULL; 
	lineStyles = NULL; 
}

ShapeWithStyle::~ShapeWithStyle() {	
	if (fillStyles != NULL) delete []fillStyles;
	if (lineStyles != NULL) delete []lineStyles;
}

ShapeWithStyle::ShapeWithStyle(ShapeWithStyle *p) : Shape(p) {
	this->numFillStyles = p->numFillStyles;
	this->numLineStyles = p->numLineStyles;
	this->fillStyles = new FILLSTYLE[p->getNumFillStyles()];
	for (int i=0; i< p->getNumFillStyles(); i++) {
		this->fillStyles[i] = p->getFillStyles()[i];
		if (p->getFillStyles()[i].getGradient() != NULL) {
			GRADIENT *a = new GRADIENT(p->getFillStyles()[i].getGradient());
			this->fillStyles[i].setGradient(a);
		}
	}
	this->lineStyles = new LINESTYLE[p->getNumLineStyles()];
	for (int i=0; i< p->getNumLineStyles(); i++) this->lineStyles[i] = p->getLineStyles()[i];
} 

Image::Image() { 
	colormap= NULL; 
	pixels = NULL; 
	alpha_buf = NULL; 
}

Image::~Image() {
	if (colormap != NULL) delete []colormap;
	if (pixels != NULL) delete[]pixels; 
	if (alpha_buf != NULL) delete []alpha_buf;
}

