/**
Timing and syncing audio and video playback
Called by as_audio, flashplayer and swf_engine.

Copyright (C) 2004  Dag Hovland (dag.hovland@student.uib.no)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#ifndef _SWF_TIMER_
#define _SWF_TIMER_

#include "sys/time.h"		// For gettimeofday
#include "errno.h"
#include "swf.hh"		// For UI16
#include <map>

/** Used to keep track of diferent states of streaming sound playback
 * FINISHED should only be set by processSoundStreamBloc
 * EXITING is set in fillAudioHandler and by any function wishing stream to exit, eg for a jump
 * PLAYING is during startup and playing of stream
*/
enum streamstate { PLAYING, EXITING, FINISHED };	

class SwfTimer {
public:
	SwfTimer(int);
	~SwfTimer();
	SwfTimer(const SwfTimer &){} 	// SwfTimer is "singleton" - no copies!
	int	getTimeOut();		// Used by players to get remaining time till next video frame should be diaplsyed
	bool	skipVideoFrame();		// Called by players. True if streaming sound is playing and video is lagging
	void	waitForAudio();		// Waits until videoFrame  is reached by sound thread or soundStreaming
	void	waitForAudioFrame(UI16, int);	// Waits until specified audio frame number is reached by sound thread or soundStreaming stopped
	void	screenUpdate();		// Called by video player to zero timeout counters
	// Called from streaming audio thread to indicate what audio frame is currently sent to soundcard
	void	setSoundFrame(UI16, int frame, int skip);		 
	int	getSoundFrame(UI16);		// Returns estimated audio frame from several parameters
	void	setVideoFrame(int f) { videoFrame = f; };
	// Called from players to see what frame is played
	int	getVideoFrame() { return videoFrame; };
	// Set by audio stream. Indicating whether audiostream playing
	void	setSoundParams(int samples, int freq, int soundStreamSampleCount);		
	void	setSoundStreamPlaying(UI16,streamstate);
	streamstate	getSoundStreamPlaying(UI16);
	void	waitForStreamEnd(UI16);
	void	stop();
	bool	getStreamIdStarted(int streamId);
	void	setStreamIdStarted(int streamId);

private:
	void lockStream();
	void lockFrame();
	void unlockStream();
	void unlockFrame();
	int	headerFrameDelay;		// The time delay between frames specified in the header
	struct timeval	lastUpdate;
	map<UI16,streamstate>	soundStreamPlaying;		// True if streaming sound is currently playing in sprite with spriteId

	map<UI16, int>	soundFrame;		// Number of lasst soundStreamBlock set by setSoundFrame, if streaming sound is playing
	struct timeval frameTime;		// Time when this number was set. Used to estimate current playing
	int	soundSkip;			// The position in the frame of start of the block, counted in samples

	int	videoFrame;		// Number of last video frame displayed with same soundStreamBlock

	int 	samples;			// Size of SDL audio buffer in samples
	int	freq;				// Frequency of sound in samples / sec
	int	soundStreamSampleCount;		// Average number of samples in each soundStreamBlockTag
	int 	streamLocks;			// Counts how many threads are waiting in waitForStreamEnd
	vector<int>	streamIds;		// Keeps track of "id"s of streams playing
};
#endif
