/**
Gplflash2 library. Unicode tools
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "common.hh"
      
/**
 * Convert wchar_t to unicode string
 **/
int wctomb (char* s, wchar_t swc) {
	wchar_t wc = swc; 
	if (s == NULL) return 0;

	size_t count;
	int result;
	if (wc < 0x80) {
		*s = (unsigned char) wc;
		return 1;
	}
	if (wc < 0x800) {
		count = 1;
		*s = (unsigned char) ((wc >> 6) | 0xC0);
	} else
#if WCHAR_T_BITS == 32
		if (wc < 0x10000)
#endif
		{
			count = 2;
			*s = (unsigned char) ((wc >> 12) | 0xE0);
		}
#if WCHAR_T_BITS == 32
		else if (wc < 0x200000) {
			count = 3;
			*s = (unsigned char) ((wc >> 18) | 0xF0);
		} else if (wc < 0x4000000) {
			count = 4;
			*s = (unsigned char) ((wc >> 24) | 0xF8);
		} else if (wc < 0x80000000) {
			count = 5;
			*s = (unsigned char) ((wc >> 30) | 0xFC);
		} else {
			return -1;
		}
#endif
	result = count + 1;
	do {
		*++s = (unsigned char)(((wc >> (6 * --count)) & 0x3F) | 0x80);
	} while (count > 0);
	return result;
}

/**
 * Convert an UTF-8 encoded char* to a wchar_t
 **/
size_t mbrtowc (wchar_t* pwc, const char* s, size_t n) {
	if (n == 0) return (size_t)(-2);

	const char* start = s;
	unsigned char c;
	wchar_t wc;
	size_t count;
	c = (unsigned char) *s;
	if (c < 0x80) { 
		if (pwc != NULL) *pwc = (wchar_t) c;
		return (c != 0);
	} else if (c < 0xC0) {
		// Spurious 10XXXXXX byte is invalid.
		goto bad_input;
	}
	s++;
	n--;
	if (c < 0xE0) {
		wc = (wchar_t)(c & 0x1F) << 6;
		count = 1;
		if (c < 0xC2) goto bad_input;
	} else if (c < 0xF0) {
		wc = (wchar_t)(c & 0x0F) << 12;
		count = 2;
	}
#if WCHAR_T_BITS == 32
	else if (c < 0xF8) {
		wc = (wchar_t)(c & 0x07) << 18;
		count = 3;
	} else if (c < 0xFC) {
		wc = (wchar_t)(c & 0x03) << 24;
		count = 4;
	} else if (c < 0xFE) {
		wc = (wchar_t)(c & 0x01) << 30;
		count = 5;
	}
#endif
	else goto bad_input;
	
	for (;;) {
		if (n == 0) return (size_t)(-2);
		n--;
		c = (unsigned char) *s++ ^ 0x80;
		if (!(c < 0x40)) goto bad_input;
		wc |= (wchar_t) c << (6 * --count);
		if (count == 0)
			break;
		if ((wchar_t) wc < ((wchar_t) 1 << (5 * count + 6)))
			goto bad_input;
	}
	if (pwc != NULL) *pwc = wc;
	return s - start;

bad_input:
  return (size_t)(-1);
}


/**
 * Print a wstring 
 **/
void printWstring(wstring w) {
	for (wstring::size_type i=0; i< w.size(); i++) {
		char s[10]; memset(s,0,10);
		wctomb(s,w.at(i));
		printf("%s",s);
	}
}

/**
 * Print a wstring 
 **/
void fprintWstring(FILE *f,wstring w) {
	for (wstring::size_type i=0; i< w.size(); i++) {
		char s[10]; memset(s,0,10);
		wctomb(s,w.at(i));
		fprintf(f,"%s",s);
	}
}

/**
 * Convert an ASCII char* to a wstring
 **/
wstring stringTowstring(const char *s) {
	int len = strlen(s);
	wchar_t *tmp = (wchar_t*)malloc((len+1)*sizeof(wchar_t));
	for (int i=0; i< len; i++)  tmp[i] = (wchar_t) (unsigned char)s[i];
	tmp[len]=0;
	wstring res = wstring(tmp);
	free(tmp);
	return res;
}

/**
 * Convert an UTF8 encoded string to wstring
 **/
wstring utf8Towstring(const char *s,int len) {
	int len0 = len;
	const char *s0 = s;
	if (len ==0) len = strlen(s);
	wstring res;
	while (len >0) {
		wchar_t w;
		int j = mbrtowc(&w,s,len);
		if (j > 0) {
			s += j;
			len -= j;
			res += w;
		} else {
			printf("utf8Towstring: decoding error! (error:%d)\n",j);  puts((char*)s0);
			// It's not UTF-8, it's extended ASCII.
			return extentedAsciiTowstring(s0,len0);
		}
	}
	return res;
}

/**
 * Convert an extended ASCII encoded string to wstring
 **/
wstring extentedAsciiTowstring(const char *s,int len) {
	if (len ==0) len = strlen(s);
	wstring res = L"";
	for (int j=0; j< len ; j++) {
		wchar_t w = (wchar_t) s[j];
		res += w;
	}
	return res;
}

/**
 * Convert a wchar_t string to an UTF8 one
 **/
void wstringToUtf8(const wstring src, char *s) {
	for (string::size_type i=0; i< src.size(); i++) {
		s += wctomb(s,src[i]);
	}
	*s = 0;
}

wstring toUpper(wstring in) {
	wstring res = L"";
	for (wstring::size_type i=0; i < in.size(); i++) {
		int k = in[i];
		if (k >= 'a' && k <= 'z') k = k -'a' + 'A';
		res += k;
	}
	return res;
}
